import requests
import json

"""
Let's start with simple function specifications:

1. if apikey is required, use "api_" before the plugin name:

	api_service.py

2. the function name must be "get" and take as parameter "domain":

	def get(domain):
		foo

3. the function must return a possibly unique list of subdomains:

	['sub1.domain.com', 'sub2.domain.com']

4. to parse the results it is recommended to use the standard modules such as:

	requests, json, bs4, re
"""

# author: Gianni Amato
# plugin: certspotter
# version: 1.0

def get(domain):
	# certspotter -> LIST -> JSON: key -> dns_names
	url = "https://api.certspotter.com/v1/issuances?domain={domain}&include_subdomains=true&expand=dns_names".format(domain=domain)
	resp = requests.get(url, timeout=5).text
	
	resp = json.loads(resp)

	result = []
	for item in resp:
		for subdomain in item['dns_names']:
			# too many dns names, filter for dns witch contains the domain
			if domain in subdomain:
				if subdomain not in result:
					result.append(subdomain)

	return result


import requests
import json

"""
Let's start with simple function specifications:

1. if apikey is required, use "api_" before the plugin name:

	api_service.py

2. the function name must be "get" and take as parameter "domain":

	def get(domain):
		foo

3. the function must return a possibly unique list of subdomains:

	['sub1.domain.com', 'sub2.domain.com']

4. to parse the results it is recommended to use the standard modules such as:

	requests, json, bs4, re
"""

# author: Gianni Amato
# plugin: alienvault
# version: 1.0

def get(domain):
	# alienvault -> JSON: key -> hostname
	url = "https://otx.alienvault.com/api/v1/indicators/domain/{domain}/passive_dns".format(domain=domain)
	resp = requests.get(url, timeout=5).text
	
	resp = json.loads(resp)
	
	result = []
	for item in resp['passive_dns']:
		subdomain = item['hostname']
		if subdomain not in result:
			result.append(subdomain)

	return result



import os
import sys
from . import report
from . import output

# extra arguments ["--report", "--plot", "--csv"]
# when domain not is a domain name but it's a command
def parse_and_exit(args):
    if len(args) == 3 and args[1] in ["--report", "--plot", "--csv"]:

        # report
        if args[1] == "--report":
            if args[2].endswith(".json"):
                if os.path.isfile(args[2]):
                    data = report.terminal(args[2])
                    if not data:
                        sys.exit("report not found or invalid json")
                    if data: sys.exit(data)
                sys.exit("report not found: %s" % args[2])
            sys.exit("try using: knockpy --report path/to/domain.com_yyyy_mm_dd_hh_mm_ss.json")

        # plot
        elif args[1] == "--plot":
            if args[2].endswith(".json"):
                if os.path.isfile(args[2]):
                    data = report.load_json(args[2])
                    if not data:
                        sys.exit("report not found or invalid json")
                    if data: 
                        plotting = report.plot(data)
                        if not plotting:
                            print("Plot needs these libraries. Use 'pip' to install them:\n- matplotlib\n- networkx\n- PyQt5")
                    sys.exit()
                sys.exit("report not found: %s" % args[2])
            sys.exit("try using: knockpy --plot path/to/domain.com_yyyy_mm_dd_hh_mm_ss.json")

        # csv
        elif args[1] == "--csv":
            if args[2].endswith(".json"):
                if os.path.isfile(args[2]):
                    data = report.load_json(args[2])
                    if data: 
                        csv_file = args[2].replace(".json", ".csv")
                        output.write_csv(csv_file, report.csv(data))
                        sys.exit("csv report: %s" % csv_file)
                sys.exit("report not found: %s" % args[2])
            sys.exit("try using: knockpy --csv path/to/domain.com_yyyy_mm_dd_hh_mm_ss.json")


import requests
import bs4

"""
Let's start with simple function specifications:

1. if apikey is required, use "api_" before the plugin name:

	api_service.py

2. the function name must be "get" and take as parameter "domain":

	def get(domain):
		foo

3. the function must return a possibly unique list of subdomains:

	['sub1.domain.com', 'sub2.domain.com']

4. to parse the results it is recommended to use the standard modules such as:

	requests, json, bs4, re
"""

# author: Gianni Amato
# plugin: rapiddns
# version: 1.0

def get(domain):
	# rapiddns -> TABLE -> scraping 'td'
	url = "https://rapiddns.io/subdomain/{domain}#result".format(domain=domain)
	resp = requests.get(url, timeout=5).text
	soup = bs4.BeautifulSoup(resp, "html.parser")

	result = []
	for item in soup.find_all("td"):
		subdomain = item.text
		if subdomain.endswith(domain):
			result.append(subdomain)

	return result


import os
import re
import time
from . import output
from importlib.machinery import SourceFileLoader

# purge wordlist
def purge(wordlist):
    return [word for word in wordlist if word and re.match("[a-z0-9\.-]", word)]

# get local dictionary
def localscan(filename):
    try:
        wlist = open(filename,'r').read().split("\n")
    except:
        if not silent_mode: output.progressPrint("wordlist not found: {filename}".format(filename=filename))
        return []
    return filter(None, wlist)

# get remote wordlist using plugin
def remotescan(domain):
    result = []

    # plugin_test is global variable
    if plugin_test:
        plugin_test_results = {}
        plugin_test_timeinit = time.time()
    
    for plugin in os.listdir(plugin_folder):
        
        # filter for .py scripts and exclude __init__.py file
        if plugin.endswith('.py') and plugin != '__init__.py':
            plugin_path = os.path.join(plugin_folder, plugin)
            
            try:
                # plugin_test is global variable
                if plugin_test:
                    plugin_test_timestart = time.time()

                if not silent_mode: 
                    output.progressPrint('') # print empty line
                    output.progressPrint(plugin) # print name of the module

                # load module
                foo = SourceFileLoader(plugin, plugin_path).load_module()
                
                # get module's result
                plugin_result = foo.get(domain)
                    
                if plugin_test:
                    # create dictionary with plugin info
                    plugin_time_elapsed = time.time() - plugin_test_timestart
                    plugin_time_elapsed = time.strftime("%H:%M:%S", time.gmtime(plugin_time_elapsed))
                    plugin_test_match = len(plugin_result)
                    plugin_test_results.update({plugin: {"time": plugin_time_elapsed, "match": plugin_test_match, "error": False}})

                # add subdomains
                result = result + plugin_result
            except:
                # print plugin error and sleep 3 secs.
                if not silent_mode: 
                    output.progressPrint("error plugin -> "+plugin)
                    time.sleep(3)
                
                if plugin_test:
                    plugin_time_elapsed = time.time() - plugin_test_timestart
                    plugin_time_elapsed = time.strftime("%H:%M:%S", time.gmtime(plugin_time_elapsed))
                    plugin_test_results.update({plugin: {"time": plugin_time_elapsed, "match": 0, "error": True}})

                continue
    
    result = list(set([r.lower() for r in result]))
    subdomains = [item.replace('.'+domain, '') for item in result]
    subdomains = purge(subdomains)
    
    # return test results
    if plugin_test:
        # add final results
        plugin_test_timeend = time.time() - plugin_test_timeinit
        plugin_test_timeend = time.strftime("%H:%M:%S", time.gmtime(plugin_test_timeend))
        plugin_test_error = [item for item in plugin_test_results.keys() if plugin_test_results[item]["error"]]
        plugin_test_list = list(plugin_test_results.keys())
        plugin_test_results.update({
            "_results": 
                {
                    "time": plugin_test_timeend,
                    "plugins": {
                        "count": len(plugin_test_list),
                        "list": plugin_test_list,
                        "error": plugin_test_error,
                        },                        
                    "subdomains": {
                        "count": len(subdomains),
                        "list": subdomains,
                        }
                }
            })
        return plugin_test_results
    
    return subdomains

# get wordlist local and/or remote
def get(domain, params):
    local, remote = [], []

    global silent_mode
    silent_mode = params["silent_mode"]
    global plugin_test
    plugin_test = params["plugin_test"]
    global local_wordlist
    local_wordlist = params["wordlist"]
    global plugin_folder
    plugin_folder = params["plugin_folder"]

    if not os.path.isfile(local_wordlist):
        return None, None

    if plugin_test:
        return remotescan(domain)

    if not params["no_local"]:
        local = list(localscan(local_wordlist))

    if not params["no_remote"]:
        remote = list(remotescan(domain))

    return local, remote


from colorama import Fore, Style
from . import request
import time
import json
import sys

# print progressbar
def progressPrint(text):
    if not text: text = " "*80
    text_dim = Style.DIM + text + Style.RESET_ALL
    sys.stdout.write("%s\r" % text_dim)
    sys.stdout.flush()
    sys.stdout.write("\r")

# colorize line
def colorizeHeader(text, count, sep):
    newText = Style.BRIGHT + Fore.YELLOW + text + Style.RESET_ALL
    _count = str(len(count)) if isinstance(count, list) else str(count)

    newCount = Style.BRIGHT + Fore.CYAN + _count + Style.RESET_ALL

    if len(count) == 0:
        newText = Style.DIM + text + Style.RESET_ALL
        newCount = Style.DIM + _count + Style.RESET_ALL
    newSep = " " + Fore.MAGENTA + sep + Style.RESET_ALL

    return newText + newCount + newSep

# print wordlist and target information
def headerPrint(local, remote, domain):
    """
    local: 0 | remote: 270
    
    Wordlist: 270 | Target: domain.com | Ip: 123.123.123.123
    """

    line = colorizeHeader("local: ", local, "| ")
    line += colorizeHeader("remote: ", remote, "\n")
    line += "\n"
    line += colorizeHeader("Wordlist: ", local + remote, "| ")

    req = request.dns(domain)
    if req != []:
        ip_req = req[2][0]
        ip = ip_req if req else ""
    else:
        ip = "None"

    line += colorizeHeader("Target: ", domain, "| ")
    line += colorizeHeader("Ip: ", ip, "\n")
    
    return line

# print header before of match-line (linePrint)
def headerBarPrint(time_start, max_len, no_http):
    """
    21:57:55

    Ip address      Subdomain               Real hostname
    --------------- ----------------------- ----------------------------
    """

    # time_start
    line = Style.BRIGHT
    line += time.strftime("%H:%M:%S", time.gmtime(time_start)) + "\n\n"

    # spaces
    spaceIp = " " * (16 - len("Ip address"))
    spaceSub = " " * ((max_len + 1) - len("Subdomain"))

    # dns only
    if no_http:
        line += "Ip address" +spaceIp+ "Subdomain" +spaceSub+ "Real hostname" + "\n"
        line += Style.RESET_ALL
        line += "-" * 15 + " " + "-" * max_len + " " + "-" * max_len
    
    # http
    else:
        spaceCode = " " * (5 - len("Code"))
        spaceServ = " " * ((max_len + 1) - len("Server"))
        line += "Ip address" +spaceIp+ "Code" +spaceCode+ "Subdomain" +spaceSub+ "Server" +spaceServ+ "Real hostname" + "\n"
        line += Style.RESET_ALL
        line += "-" * 15 + " " + "-" * 4 + " " + "-" * max_len + " " + "-" * max_len + " " + "-" * max_len
    
    return line

# change json for different scan: dns or dns + http
def jsonizeRequestData(req, target):
    if len(req) == 3:
        subdomain, aliasList, ipList = req
        domain = subdomain if subdomain != target else ""

        data = {
            "target": target,
            "domain": domain,
            "alias": aliasList,
            "ipaddr": ipList
            }
    elif len(req) == 5:
        subdomain, aliasList, ipList, code, server = req
        domain = subdomain if subdomain != target else ""

        data = {
            "target": target,
            "domain": domain,
            "alias": aliasList,
            "ipaddr": ipList,
            "code": code,
            "server": server
            }
    else:
        data = {}

    return data

# print match-line while it's working
def linePrint(data, max_len):
    """
    123.123.123.123   click.domain.com     click.virt.s6.exactdomain.com
    """ 

    # just a fix, print space if not domain
    _domain = " "*max_len if not data["domain"] else data["domain"]

    # case dns only
    if len(data.keys()) == 4:
        spaceIp = " " * (16 - len(data["ipaddr"][0]))
        spaceSub = " " * ((max_len + 1) - len(data["target"]))
        _target = Style.BRIGHT + Fore.CYAN + data["target"] + Style.RESET_ALL if data["alias"] else data["target"]
        line = data["ipaddr"][0] +spaceIp+ _target +spaceSub+ _domain
    
    # case dns +http
    elif len(data.keys()) == 6:
        data["server"] = data["server"][:max_len]

        spaceIp = " " * (16 - len(data["ipaddr"][0]))
        spaceSub = " " * ((max_len + 1) - len(data["target"]))
        spaceCode = " " * (5 - len(str(data["code"])))
        spaceServer = " " * ((max_len + 1) - len(data["server"]))
        
        if data["code"] == 200:
            _code = Style.BRIGHT + Fore.GREEN + str(data["code"]) + Style.RESET_ALL
            _target = Style.BRIGHT + Fore.GREEN + data["target"] + Style.RESET_ALL
        elif str(data["code"]).startswith("4"):
            _code = Style.BRIGHT + Fore.MAGENTA + str(data["code"]) + Style.RESET_ALL
            _target = Style.BRIGHT + Fore.MAGENTA + data["target"] + Style.RESET_ALL
        elif str(data["code"]).startswith("5"):
            _code = Style.BRIGHT + Fore.RED + str(data["code"]) + Style.RESET_ALL
            _target = Style.BRIGHT + Fore.RED + data["target"] + Style.RESET_ALL
        else:
            _code = str(data["code"])
            _target = Style.BRIGHT + Fore.CYAN + data["target"] + Style.RESET_ALL if data["domain"] else data["target"]

        line = data["ipaddr"][0] +spaceIp+ _code +spaceCode+ _target +spaceSub+ data["server"] +spaceServer+ _domain

    return line

# print footer at the end after match-line (linePrint)
def footerPrint(time_end, time_start, results):
    """
    21:58:06

    Ip address: 122 | Subdomain: 93 | elapsed time: 00:00:11 
    """

    progressPrint("")
    elapsed_time = time_end - time_start
    line = Style.BRIGHT
    line += "\n"
    line += time.strftime("%H:%M:%S", time.gmtime(time_end))
    line += "\n\n"
    line += Style.RESET_ALL

    ipList = []
    for i in results.keys():
        for ii in results[i]["ipaddr"]:
            ipList.append(ii)

    line += colorizeHeader("Ip address: ", list(set(ipList)), "| ")
    line += colorizeHeader("Subdomain: ", list(results.keys()), "| ")
    line += colorizeHeader("elapsed time: ", time.strftime("%H:%M:%S", time.gmtime(elapsed_time)), "\n")

    return line

# create json file
def write_json(path, json_data):
    f = open(path, "w")
    f.write(json.dumps(json_data, indent=4))
    f.close()

# create csv file
def write_csv(path, csv_data):
    f = open(path, "w")
    f.write(csv_data)
    f.close()


from . import output
from . import request

def start(max_len, domain, subdomain, percentage, results, params):
    ctrl_c = "(ctrl+c) | "

    #output.progressPrint(ctrl_c + subdomain)
    target = subdomain+"."+domain
    if not params["silent_mode"]: output.progressPrint(ctrl_c + str(percentage*100)[:4] + "% | " + target + " "*max_len)
    req = request.dns(target, params["dns"])

    if not req: return None

    req = list(req)
    ip_req = req[2][0]

    if ip_req in params["no_ip"]: return None

    # dns only
    if params["no_http"]:
        # print line and update report
        data = output.jsonizeRequestData(req, target)
        if not params["silent_mode"]: print (output.linePrint(data, max_len))
        del data["target"]
        return results.update({target: data})

    # dns and http(s)
    https = request.https(target, params["useragent"])
    
    if https:
        for item in https:
            req.append(item)
    else:
        http = request.http(target, params["useragent"])
        
        if http:
            for item in http:
                req.append(item)
        else:
            req.append("")
            req.append("")

    # print line and update report
    data = output.jsonizeRequestData(req, target)
    if data["code"] in params["no_http_code"]: return None
    if not params["silent_mode"]: print (output.linePrint(data, max_len))
    del data["target"]
    
    return results.update({target: data})


import requests
import re

"""
Let's start with simple function specifications:

1. if apikey is required, use "api_" before the plugin name:

	api_service.py

2. the function name must be "get" and take as parameter "domain":

	def get(domain):
		foo

3. the function must return a possibly unique list of subdomains:

	['sub1.domain.com', 'sub2.domain.com']

4. to parse the results it is recommended to use the standard modules such as:

	requests, json, bs4, re
"""

# author: Gianni Amato
# plugin: archive
# version: 1.0

def get(domain):
	# webarchive -> TEXT URL LIST -> match subdomain
	url = "https://web.archive.org/cdx/search/cdx?url=*.{domain}/*&output=txt&fl=original&collapse=urlkey".format(domain=domain)
	resp = requests.get(url, timeout=5).text

	result = []
	pattern = "http(s)?:\/\/(.*\.%s)" % domain
	for item in resp.split('\n'):
		match = re.match(pattern, item)

		if match and re.match("^[a-zA-Z0-9-\.]*$", match.groups()[1]):
			subdomain = match.groups()[1]
			if subdomain not in result:
				result.append(subdomain)
	
	return result



import os
import time
import json
from . import output

user_agent = [
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.14; rv:65.0) Gecko/20100101 Firefox/65.0",
    "Mozilla/5.0 (MSIE 10.0; Windows NT 6.1; Trident/5.0)",
    "Opera/9.80 (Windows NT 6.0) Presto/2.12.388 Version/12.14",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_3) AppleWebKit/537.75.14 (KHTML, like Gecko) Version/7.0.3 Safari/7046A194A"
]

# import json file
def load_json(report):
    try:
        report_json = json.load(open(report))
        del report_json["_meta"]
        return report_json
    except:
        return None

# save output and add _meta to json file
def save(results, domain, time_start, time_end, len_wordlist, version, output_folder):
    _meta = {
        "name": "knockpy",
        "version": version,
        "time_start": time_start,
        "time_end": time_end,
        "domain": domain,
        "wordlist": len_wordlist
        }
    
    results.update({"_meta": _meta})
    strftime = "%Y_%m_%d_%H_%M_%S"
    date = time.strftime(strftime, time.gmtime(time_end)) 
    path = output_folder + os.sep + domain + "_" + date + ".json"
    output.write_json(path, results)

# convert json to csv
def csv(report):
    csv_data = ""
    for item in report.keys():
        if len(report[item]) == 5:
            """
            fix injection:
            https://github.com/guelfoweb/knock/commit/156378d97f10871d30253eeefe15ec399aaa0b03
            https://www.exploit-db.com/exploits/49342
            """
            csv_injection = ("+", "-", "=", "@")
            if report[item]["server"].startswith(csv_injection):
                report[item]["server"] = "'" + report[item]["server"]
            
            csv_data += "%s;%s;%s;%s;%s" % (report[item]["ipaddr"][0],
                                    report[item]["code"],
                                    item,
                                    report[item]["server"],
                                    report[item]["domain"])
        if len(report[item]) == 3:
            csv_data += "%s;%s;%s" % (report[item]["ipaddr"][0],
                                    item,
                                    report[item]["domain"])
        csv_data += "\n"
    return csv_data

# convert json to human text to show in terminal
def terminal(domain):
    report_json = load_json(domain)
    
    # report not found or invalid json
    if report_json == None: 
        return None

    results = ""
    for item in report_json.keys():
        report_json[item].update({"target": item})
        max_len = len(max(list(report_json.keys()), key=len))
        results += output.linePrint(report_json[item], max_len) + "\n"
    return results

# plotting relationships
def plot(report):
    # todo:
    # get modules list from sys.modules.keys()
    try:
        import matplotlib.pyplot as plt
        import networkx as nx
    except:
        return None


    dataset = []
    for item in report.keys():
        dataset.append((report[item]["ipaddr"][0], item))

    g = nx.Graph()
    g.add_edges_from(dataset)

    pos = nx.spring_layout(g)
    nx.draw(g, pos, node_size=50, node_color="r", edge_color="c", with_labels=True, width=0.7, alpha=0.9)
    plt.show()